package dev.pfaff.glass.apple.gcd;

import org.jspecify.annotations.Nullable;

import java.lang.foreign.Arena;
import java.lang.foreign.Linker;
import java.lang.foreign.MemorySegment;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionException;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.function.BooleanSupplier;

import static dev.pfaff.glass.apple.gcd.GrandCentralDispatch.*;

/**
 * Note that {@link #dispatchSync(Runnable)} and friends use
 * <a href="https://developer.apple.com/documentation/dispatch/3191902-dispatch_async_and_wait_f">dispatch_async_and_wait_f</a>,
 * not <a href="https://developer.apple.com/documentation/dispatch/1453123-dispatch_sync_f">dispatch_sync_f</a>.
 */
public final class DispatchQueue {
	private final MemorySegment MS_queue;
	private final MemorySegment MS_runNextTask;
	@SuppressWarnings("FieldCanBeLocal")
	private final Arena arena;
	private final ConcurrentLinkedQueue<Runnable> queue = new ConcurrentLinkedQueue<>();
	private final BooleanSupplier shouldSkipQueue;

	public DispatchQueue(MemorySegment MS_queue, Arena arena, BooleanSupplier shouldSkipQueue) {
		this.MS_queue = Objects.requireNonNull(MS_queue);
		this.arena = Objects.requireNonNull(arena);
		this.MS_runNextTask = Linker.nativeLinker().upcallStub(MH_runNextTask.bindTo(queue), FD_runTaskBound, arena);
		this.shouldSkipQueue = shouldSkipQueue;
	}

	public DispatchQueue(MemorySegment MS_queue, Arena arena) {
		this(MS_queue, arena, () -> false);
	}

	public CompletableFuture<@Nullable Void> dispatchAsync(Runnable task) {
		var fut = new CompletableFuture<@Nullable Void>();
		dispatchAsyncRaw(() -> {
			try {
				task.run();
				fut.complete(null);
			} catch (Throwable e) {
				fut.completeExceptionally(e);
			}
		});
		return fut;
	}

	public <T> CompletableFuture<T> dispatchAsync(Callable<T> task) {
		var fut = new CompletableFuture<T>();
		dispatchAsyncRaw(() -> {
			try {
				fut.complete(task.call());
			} catch (Throwable e) {
				fut.completeExceptionally(e);
			}
		});
		return fut;
	}

	public void dispatchSync(Runnable task) {
		var runnable = new Runnable() {
			@Nullable
			Throwable exception;

			@Override
			public void run() {
				try {
					task.run();
				} catch (Throwable e) {
					exception = e;
				}
			}
		};
		dispatchSyncRaw(runnable);
		if (runnable.exception != null) throw new CompletionException(runnable.exception);
	}

	public <T> T dispatchSync(Callable<T> task) {
		var runnable = new Runnable() {
			T value;
			@Nullable
			Throwable exception;

			@Override
			public void run() {
				try {
					value = task.call();
				} catch (Throwable e) {
					exception = e;
				}
			}
		};
		dispatchSyncRaw(runnable);
		if (runnable.exception != null) throw new CompletionException(runnable.exception);
		return runnable.value;
	}

	private void dispatchAsyncRaw(Runnable runnable) {
		if (shouldSkipQueue.getAsBoolean()) {
			runnable.run();
			return;
		}
		queue.add(runnable);

		try {
			MH_dispatch_async_f.invokeExact(MS_queue, MemorySegment.NULL, MS_runNextTask);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	private void dispatchSyncRaw(Runnable runnable) {
		if (shouldSkipQueue.getAsBoolean()) {
			runnable.run();
			return;
		}

		queue.add(runnable);

		try {
			MH_dispatch_async_and_wait_f.invokeExact(MS_queue, MemorySegment.NULL, MS_runNextTask);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}
}
