package dev.pfaff.glass.glfw;

import dev.pfaff.glass.glfw.callbacks.ErrorCallback;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.*;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;

import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodType.methodType;

public final class GLFW {
	static final Linker LINKER = Linker.nativeLinker();

	private static final MethodHandle MH_errorCallbackImpl;
	public static final FunctionDescriptor FD_GLFWerrorfun = FunctionDescriptor.ofVoid(JAVA_INT,
			ADDRESS.withTargetLayout(JAVA_BYTE));

	private static Arena errorCallbackArena;

	static {
		try {
			MH_errorCallbackImpl = MethodHandles.lookup().findStatic(GLFW.class,
					"errorCallbackImpl",
					methodType(void.class, ErrorCallback.class, int.class, MemorySegment.class));
		} catch (NoSuchMethodException | IllegalAccessException e) {
			throw new RuntimeException(e);
		}
	}

	private static void errorCallbackImpl(ErrorCallback callback, int code, MemorySegment descriptionPtr) {
		var description = descriptionPtr.reinterpret(Long.MAX_VALUE).getString(0, StandardCharsets.UTF_8);
		callback.call(code, description);
	}

	public static final AddressLayout ML_GLFWwindow = ADDRESS.withName("GLFWwindow*");
	public static final AddressLayout ML_GLFWmonitor = ADDRESS.withName("GLFWmonitor*");
	public static final AddressLayout ML_GLFWerrorfun = ADDRESS.withName("GLFWerrorfun");

	public static final MethodHandle MH_glfwInit = LINKER.downcallHandle(FunctionDescriptor.of(JAVA_BOOLEAN));
	public static final MethodHandle MH_glfwTerminate = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwPollEvents = LINKER.downcallHandle(FunctionDescriptor.ofVoid());
	public static final MethodHandle MH_glfwSetErrorCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWerrorfun,
			ML_GLFWerrorfun));

	private final MemorySegment MS_glfwInit;
	private final MemorySegment MS_glfwTerminate;
	private final MemorySegment MS_glfwPollEvents;
	private final MemorySegment MS_glfwSetErrorCallback;

	final MemorySegment MS_glfwMakeContextCurrent;
	final MemorySegment MS_glfwGetCurrentContext;
	final MemorySegment MS_glfwSwapInterval;
	final MemorySegment MS_glfwExtensionSupported;
	final MemorySegment MS_glfwCreateWindow;
	final MemorySegment MS_glfwWindowHint;
	final MemorySegment MS_glfwDestroyWindow;
	final MemorySegment MS_glfwWindowShouldClose;
	final MemorySegment MS_glfwGetFramebufferSize;
	final MemorySegment MS_glfwGetWindowSize;
	final MemorySegment MS_glfwGetWindowContentScale;
	final MemorySegment MS_glfwSwapBuffers;
	final MemorySegment MS_glfwShowWindow;
	final MemorySegment MS_glfwHideWindow;
	final MemorySegment MS_glfwSetWindowSize;

	final MemorySegment MS_glfwSetCharCallback;
	final MemorySegment MS_glfwSetCursorPosCallback;
	final MemorySegment MS_glfwSetKeyCallback;
	final MemorySegment MS_glfwSetMouseButtonCallback;
	final MemorySegment MS_glfwSetDropCallback;
	final MemorySegment MS_glfwSetScrollCallback;
	final MemorySegment MS_glfwSetWindowPosCallback;
	final MemorySegment MS_glfwSetWindowSizeCallback;
	final MemorySegment MS_glfwSetWindowCloseCallback;
	final MemorySegment MS_glfwSetWindowRefreshCallback;
	final MemorySegment MS_glfwSetWindowFocusCallback;
	final MemorySegment MS_glfwSetWindowIconifyCallback;
	final MemorySegment MS_glfwSetWindowMaximizeCallback;
	final MemorySegment MS_glfwSetFramebufferSizeCallback;
	final MemorySegment MS_glfwSetWindowContentScaleCallback;

	public GLFW(SymbolLookup symbolLookup) {
		this.MS_glfwInit = symbolLookup.find("glfwInit").get();
		this.MS_glfwTerminate = symbolLookup.find("glfwTerminate").get();
		this.MS_glfwPollEvents = symbolLookup.find("glfwPollEvents").get();
		this.MS_glfwSetErrorCallback = symbolLookup.find("glfwSetErrorCallback").get();

		this.MS_glfwMakeContextCurrent = symbolLookup.find("glfwMakeContextCurrent").get();
		this.MS_glfwGetCurrentContext = symbolLookup.find("glfwGetCurrentContext").get();
		this.MS_glfwSwapInterval = symbolLookup.find("glfwSwapInterval").get();
		this.MS_glfwExtensionSupported = symbolLookup.find("glfwExtensionSupported").get();
		this.MS_glfwCreateWindow = symbolLookup.find("glfwCreateWindow").get();
		this.MS_glfwWindowHint = symbolLookup.find("glfwWindowHint").get();
		this.MS_glfwDestroyWindow = symbolLookup.find("glfwDestroyWindow").get();
		this.MS_glfwWindowShouldClose = symbolLookup.find("glfwWindowShouldClose").get();
		this.MS_glfwGetFramebufferSize = symbolLookup.find("glfwGetFramebufferSize").get();
		this.MS_glfwGetWindowSize = symbolLookup.find("glfwGetWindowSize").get();
		this.MS_glfwGetWindowContentScale = symbolLookup.find("glfwGetWindowContentScale").get();
		this.MS_glfwSwapBuffers = symbolLookup.find("glfwSwapBuffers").get();
		this.MS_glfwShowWindow = symbolLookup.find("glfwShowWindow").get();
		this.MS_glfwHideWindow = symbolLookup.find("glfwHideWindow").get();
		this.MS_glfwSetWindowSize = symbolLookup.find("glfwSetWindowSize").get();

		this.MS_glfwSetCharCallback = symbolLookup.find("glfwSetCharCallback").get();
		this.MS_glfwSetCursorPosCallback = symbolLookup.find("glfwSetCursorPosCallback").get();
		this.MS_glfwSetKeyCallback = symbolLookup.find("glfwSetKeyCallback").get();
		this.MS_glfwSetMouseButtonCallback = symbolLookup.find("glfwSetMouseButtonCallback").get();
		this.MS_glfwSetDropCallback = symbolLookup.find("glfwSetDropCallback").get();
		this.MS_glfwSetScrollCallback = symbolLookup.find("glfwSetScrollCallback").get();
		this.MS_glfwSetWindowPosCallback = symbolLookup.find("glfwSetWindowPosCallback").get();
		this.MS_glfwSetWindowSizeCallback = symbolLookup.find("glfwSetWindowSizeCallback").get();
		this.MS_glfwSetWindowCloseCallback = symbolLookup.find("glfwSetWindowCloseCallback").get();
		this.MS_glfwSetWindowRefreshCallback = symbolLookup.find("glfwSetWindowRefreshCallback").get();
		this.MS_glfwSetWindowFocusCallback = symbolLookup.find("glfwSetWindowFocusCallback").get();
		this.MS_glfwSetWindowIconifyCallback = symbolLookup.find("glfwSetWindowIconifyCallback").get();
		this.MS_glfwSetWindowMaximizeCallback = symbolLookup.find("glfwSetWindowMaximizeCallback").get();
		this.MS_glfwSetFramebufferSizeCallback = symbolLookup.find("glfwSetFramebufferSizeCallback").get();
		this.MS_glfwSetWindowContentScaleCallback = symbolLookup.find("glfwSetWindowContentScaleCallback").get();
	}

	// TODO: error callback

	/**
	 * Initializes GLFW.
	 */
	public void init() {
		boolean result;
		try {
			result = (boolean) MH_glfwInit.invokeExact(MS_glfwInit);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (!result) {
			throw new RuntimeException("Failed to initialize GLFW");
		}
	}

	/**
	 * Deinitializes GLFW. You must call {@link #init()} after this call to use the library again.
	 */
	public void terminate() {
		try {
			MH_glfwTerminate.invokeExact(MS_glfwTerminate);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	public void pollEvents() {
		try {
			MH_glfwPollEvents.invokeExact(MS_glfwPollEvents);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
	}

	static <T> Arena setCallback(@NonNull MethodHandle MH_setCallback,
								 @NonNull MemorySegment MS_setCallback,
								 @NonNull T callback,
								 @NonNull MethodHandle invokerStub,
								 @NonNull FunctionDescriptor invokerFunctionDescriptor,
								 @Nullable Arena oldArena) {
		Arena newArena = Arena.ofShared();
		try {
			var stub = LINKER.upcallStub(invokerStub.bindTo(callback), invokerFunctionDescriptor, newArena);
			var _ = (MemorySegment) MH_setCallback.invokeExact(MS_setCallback, stub);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (oldArena != null) oldArena.close();
		return newArena;
	}

	public void setErrorCallback(ErrorCallback callback) {
		errorCallbackArena = setCallback(MH_glfwSetErrorCallback,
				MS_glfwSetErrorCallback,
				callback,
				MH_errorCallbackImpl,
				FD_GLFWerrorfun,
				errorCallbackArena);
	}
}
