package dev.pfaff.glass.glfw;

import dev.pfaff.glass.glfw.callbacks.*;
import org.jspecify.annotations.NonNull;
import org.jspecify.annotations.Nullable;

import java.lang.foreign.AddressLayout;
import java.lang.foreign.Arena;
import java.lang.foreign.FunctionDescriptor;
import java.lang.foreign.MemorySegment;
import java.lang.invoke.MethodHandle;
import java.lang.invoke.MethodHandles;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.util.Objects;
import java.util.stream.StreamSupport;

import static dev.pfaff.glass.glfw.GLFW.*;
import static java.lang.foreign.ValueLayout.*;
import static java.lang.invoke.MethodHandles.insertArguments;

/**
 * Keeps a reference to the callbacks alive.
 */
public final class WindowCallbackManager {
	private static final AddressLayout ML_GLFWcharfun = ADDRESS.withName("GLFWcharfun");
	private static final AddressLayout ML_GLFWcursorposfun = ADDRESS.withName("GLFWcursorposfun");
	private static final AddressLayout ML_GLFWkeyfun = ADDRESS.withName("GLFWkeyfun");
	private static final AddressLayout ML_GLFWmousebuttonfun = ADDRESS.withName("GLFWmousebuttonfun");
	private static final AddressLayout ML_GLFWdropfun = ADDRESS.withName("GLFWdropfun");
	private static final AddressLayout ML_GLFWscrollfun = ADDRESS.withName("GLFWscrollfun");
	private static final AddressLayout ML_GLFWframebuffersizefun = ADDRESS.withName("GLFWframebuffersizefun");
	private static final AddressLayout ML_GLFWwindowposfun = ADDRESS.withName("ML_GLFWwindowposfun");
	private static final AddressLayout ML_GLFWwindowsizefun = ADDRESS.withName("GLFWwindowsizefun");
	private static final AddressLayout ML_GLFWwindowclosefun = ADDRESS.withName("GLFWwindowclosefun");
	private static final AddressLayout ML_GLFWwindowrefreshfun = ADDRESS.withName("GLFWwindowrefreshfun");
	private static final AddressLayout ML_GLFWwindowfocusfun = ADDRESS.withName("GLFWwindowfocusfun");
	private static final AddressLayout ML_GLFWwindowiconifyfun = ADDRESS.withName("GLFWwindowiconifyfun");
	private static final AddressLayout ML_GLFWwindowmaximizefun = ADDRESS.withName("GLFWwindowmaximizefun");
	private static final AddressLayout ML_GLFWwindowcontentscalefun = ADDRESS.withName("GLFWwindowcontentscalefun");

	private static final MethodHandle MH_glfwSetCharCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWcharfun,
			ML_GLFWwindow,
			ML_GLFWcharfun));
	private static final MethodHandle MH_glfwSetCursorPosCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWcursorposfun,
			ML_GLFWwindow,
			ML_GLFWcursorposfun));
	private static final MethodHandle MH_glfwSetKeyCallback =
			LINKER.downcallHandle(FunctionDescriptor.of(ML_GLFWkeyfun,
			ML_GLFWwindow,
			ML_GLFWkeyfun));
	private static final MethodHandle MH_glfwSetMouseButtonCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWmousebuttonfun,
			ML_GLFWwindow,
			ML_GLFWmousebuttonfun));
	private static final MethodHandle MH_glfwSetDropCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWdropfun,
			ML_GLFWwindow,
			ML_GLFWdropfun));
	private static final MethodHandle MH_glfwSetScrollCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWscrollfun,
			ML_GLFWwindow,
			ML_GLFWscrollfun));
	private static final MethodHandle MH_glfwSetWindowPosCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowposfun,
			ML_GLFWwindow,
			ML_GLFWwindowposfun));
	private static final MethodHandle MH_glfwSetWindowSizeCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowsizefun,
			ML_GLFWwindow,
			ML_GLFWwindowsizefun));
	private static final MethodHandle MH_glfwSetWindowCloseCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowclosefun,
			ML_GLFWwindow,
			ML_GLFWwindowclosefun));
	private static final MethodHandle MH_glfwSetWindowRefreshCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowrefreshfun,
			ML_GLFWwindow,
			ML_GLFWwindowrefreshfun));
	private static final MethodHandle MH_glfwSetWindowFocusCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowfocusfun,
			ML_GLFWwindow,
			ML_GLFWwindowfocusfun));
	private static final MethodHandle MH_glfwSetWindowIconifyCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowiconifyfun,
			ML_GLFWwindow,
			ML_GLFWwindowiconifyfun));
	private static final MethodHandle MH_glfwSetWindowMaximizeCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowmaximizefun,
			ML_GLFWwindow,
			ML_GLFWwindowmaximizefun));
	private static final MethodHandle MH_glfwSetFramebufferSizeCallback = LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWframebuffersizefun,
			ML_GLFWwindow,
			ML_GLFWframebuffersizefun));
	private static final MethodHandle MH_glfwSetWindowContentScaleCallback =
			LINKER.downcallHandle(FunctionDescriptor.of(
			ML_GLFWwindowcontentscalefun,
			ML_GLFWwindow,
			ML_GLFWwindowcontentscalefun));

	public static final FunctionDescriptor FD_GLFWcharfun = FunctionDescriptor.ofVoid(ML_GLFWwindow, JAVA_INT);
	public static final FunctionDescriptor FD_GLFWcursorposfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_DOUBLE,
			JAVA_DOUBLE);
	public static final FunctionDescriptor FD_GLFWkeyfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			JAVA_INT,
			JAVA_INT,
			JAVA_INT);
	public static final FunctionDescriptor FD_GLFWmousebuttonfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			JAVA_INT,
			JAVA_INT);
	public static final FunctionDescriptor FD_GLFWdropfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			ADDRESS.withTargetLayout(ADDRESS.withTargetLayout(JAVA_BYTE)));
	public static final FunctionDescriptor FD_GLFWscrollfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_DOUBLE,
			JAVA_DOUBLE);
	public static final FunctionDescriptor FD_GLFWwindowposfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			JAVA_INT);
	public static final FunctionDescriptor FD_GLFWwindowsizefun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			JAVA_INT);
	public static final FunctionDescriptor FD_GLFWwindowclosefun = FunctionDescriptor.ofVoid(ML_GLFWwindow);
	public static final FunctionDescriptor FD_GLFWwindowrefreshfun = FunctionDescriptor.ofVoid(ML_GLFWwindow);
	public static final FunctionDescriptor FD_GLFWwindowfocusfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_BOOLEAN);
	public static final FunctionDescriptor FD_GLFWwindowiconifyfun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_BOOLEAN);
	public static final FunctionDescriptor FD_GLFWwindowmaximizefun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_BOOLEAN);
	public static final FunctionDescriptor FD_GLFWframebuffersizefun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_INT,
			JAVA_INT);
	public static final FunctionDescriptor FD_GLFWwindowcontentscalefun = FunctionDescriptor.ofVoid(ML_GLFWwindow,
			JAVA_FLOAT,
			JAVA_FLOAT);

	private static final MethodHandles.Lookup MH_LOOKUP = MethodHandles.lookup();

	private static MethodHandle findCallbackImpl(Class<?> callbackType, FunctionDescriptor FD_callback) {
		try {
			return MH_LOOKUP.findStatic(WindowCallbackManager.class,
					"callbackImpl",
					FD_callback.toMethodType().insertParameterTypes(0, GLFW.class, callbackType));
		} catch (ReflectiveOperationException e) {
			throw new RuntimeException(e);
		}
	}

	private static final MethodHandle
			MH_charCallbackImpl = findCallbackImpl(CharCallback.class, FD_GLFWcharfun),
			MH_cursorPosCallbackImpl = findCallbackImpl(CursorPosCallback.class, FD_GLFWcursorposfun),
			MH_keyCallbackImpl = findCallbackImpl(KeyCallback.class, FD_GLFWkeyfun),
			MH_mouseButtonCallbackImpl = findCallbackImpl(MouseButtonCallback.class, FD_GLFWmousebuttonfun),
			MH_pathDropCallbackImpl = findCallbackImpl(PathDropCallback.class, FD_GLFWdropfun),
			MH_scrollCallbackImpl = findCallbackImpl(ScrollCallback.class, FD_GLFWscrollfun),
			MH_windowPosCallbackImpl = findCallbackImpl(WindowPosCallback.class, FD_GLFWwindowposfun),
			MH_windowSizeCallbackImpl = findCallbackImpl(WindowSizeCallback.class, FD_GLFWwindowsizefun),
			MH_windowCloseCallbackImpl = findCallbackImpl(WindowCloseCallback.class, FD_GLFWwindowclosefun),
			MH_windowRefreshCallbackImpl = findCallbackImpl(WindowRefreshCallback.class, FD_GLFWwindowrefreshfun),
			MH_windowFocusCallbackImpl = findCallbackImpl(WindowFocusCallback.class, FD_GLFWwindowfocusfun),
			MH_windowIconifyCallbackImpl = findCallbackImpl(WindowIconifyCallback.class, FD_GLFWwindowiconifyfun),
			MH_windowMaximizeCallbackImpl = findCallbackImpl(WindowMaximizeCallback.class, FD_GLFWwindowmaximizefun),
			MH_framebufferSizeCallbackImpl = findCallbackImpl(FramebufferSizeCallback.class,
					FD_GLFWframebuffersizefun),
			MH_windowContentScaleCallbackImpl = findCallbackImpl(WindowContentScaleCallback.class,
					FD_GLFWwindowcontentscalefun);

	private final GLFWwindow window;
	private Arena charCallbackArena, cursorPosCallbackArena, keyCallbackArena, mouseButtonCallbackArena,
			pathDropCallbackArena, scrollCallbackArena, windowPosCallbackArena, windowSizeCallbackArena,
			windowCloseCallbackArena, windowRefreshCallbackArena, windowFocusCallbackArena, windowIconifyCallbackArena
			, windowMaximizeCallbackArena, framebufferSizeCallbackArena, windowContentScaleCallbackArena;

	public WindowCallbackManager(GLFWwindow window) {
		this.window = Objects.requireNonNull(window);
	}

	private static void callbackImpl(GLFW glfw, CharCallback callback, MemorySegment window, int codePoint) {
		callback.call(new GLFWwindow(glfw, window), codePoint);
	}

	private static void callbackImpl(GLFW glfw, CursorPosCallback callback, MemorySegment window, double xPos, double yPos) {
		callback.call(new GLFWwindow(glfw, window), xPos, yPos);
	}

	private static void callbackImpl(GLFW glfw, KeyCallback callback, MemorySegment window, int key, int scanCode, int action,
									 int modifiers) {
		callback.call(new GLFWwindow(glfw, window), key, scanCode, action, modifiers);
	}

	private static void callbackImpl(GLFW glfw, MouseButtonCallback callback, MemorySegment window, int button, int action,
									 int modifiers) {
		callback.call(new GLFWwindow(glfw, window), button, action, modifiers);
	}

	private static void callbackImpl(GLFW glfw, PathDropCallback callback, MemorySegment window, int pathCount,
									 MemorySegment pathsPtr) {
		var paths = StreamSupport.stream(pathsPtr.spliterator(ADDRESS.withTargetLayout(JAVA_BYTE)), true)
				.map(ms -> Path.of(ms.getString(0, StandardCharsets.UTF_8)))
				.toList();
		callback.call(new GLFWwindow(glfw, window), paths);
	}

	private static void callbackImpl(GLFW glfw, ScrollCallback callback, MemorySegment window, double xOffset, double yOffset) {
		callback.call(new GLFWwindow(glfw, window), xOffset, yOffset);
	}

	private static void callbackImpl(GLFW glfw, WindowPosCallback callback, MemorySegment window, int x, int y) {
		callback.call(new GLFWwindow(glfw, window), x, y);
	}

	private static void callbackImpl(GLFW glfw, WindowSizeCallback callback, MemorySegment window, int width, int height) {
		callback.call(new GLFWwindow(glfw, window), width, height);
	}

	private static void callbackImpl(GLFW glfw, WindowCloseCallback callback, MemorySegment window) {
		callback.call(new GLFWwindow(glfw, window));
	}

	private static void callbackImpl(GLFW glfw, WindowRefreshCallback callback, MemorySegment window) {
		callback.call(new GLFWwindow(glfw, window));
	}

	private static void callbackImpl(GLFW glfw, WindowFocusCallback callback, MemorySegment window, boolean focused) {
		callback.call(new GLFWwindow(glfw, window), focused);
	}

	private static void callbackImpl(GLFW glfw, WindowIconifyCallback callback, MemorySegment window, boolean iconified) {
		callback.call(new GLFWwindow(glfw, window), iconified);
	}

	private static void callbackImpl(GLFW glfw, WindowMaximizeCallback callback, MemorySegment window, boolean maximized) {
		callback.call(new GLFWwindow(glfw, window), maximized);
	}

	private static void callbackImpl(GLFW glfw, FramebufferSizeCallback callback, MemorySegment window, int width, int height) {
		callback.call(new GLFWwindow(glfw, window), width, height);
	}

	private static void callbackImpl(GLFW glfw, WindowContentScaleCallback callback, MemorySegment window, float x, float y) {
		callback.call(new GLFWwindow(glfw, window), x, y);
	}

	private <T> Arena setCallback(@NonNull MethodHandle MH_setCallback,
								  @NonNull MemorySegment MS_setCallback,
								  @NonNull T callback,
								  @NonNull MethodHandle invokerStub,
								  @NonNull FunctionDescriptor invokerFunctionDescriptor,
								  @Nullable Arena oldArena) {
		Arena newArena = Arena.ofShared();
		try {
			var stub = LINKER.upcallStub(insertArguments(invokerStub, 0, window.glfw(), callback),
					invokerFunctionDescriptor,
					newArena);
			var _ = (MemorySegment) MH_setCallback.invokeExact(MS_setCallback, window.handle(), stub);
		} catch (Throwable e) {
			throw new RuntimeException(e);
		}
		if (oldArena != null) oldArena.close();
		return newArena;
	}

	public void setCharCallback(CharCallback callback) {
		charCallbackArena = setCallback(MH_glfwSetCharCallback,
				window.glfw().MS_glfwSetCharCallback,
				callback,
				MH_charCallbackImpl,
				FD_GLFWcharfun,
				charCallbackArena);
	}

	public void setCursorPosCallback(CursorPosCallback callback) {
		cursorPosCallbackArena = setCallback(MH_glfwSetCursorPosCallback,
				window.glfw().MS_glfwSetCursorPosCallback,
				callback,
				MH_cursorPosCallbackImpl,
				FD_GLFWcursorposfun,
				cursorPosCallbackArena);
	}

	public void setKeyCallback(KeyCallback callback) {
		keyCallbackArena = setCallback(MH_glfwSetKeyCallback,
				window.glfw().MS_glfwSetKeyCallback,
				callback,
				MH_keyCallbackImpl,
				FD_GLFWkeyfun,
				keyCallbackArena);
	}

	public void setMouseButtonCallback(MouseButtonCallback callback) {
		mouseButtonCallbackArena = setCallback(MH_glfwSetMouseButtonCallback,
				window.glfw().MS_glfwSetMouseButtonCallback,
				callback,
				MH_mouseButtonCallbackImpl,
				FD_GLFWmousebuttonfun,
				mouseButtonCallbackArena);
	}

	public void setPathDropCallback(PathDropCallback callback) {
		pathDropCallbackArena = setCallback(MH_glfwSetDropCallback,
				window.glfw().MS_glfwSetDropCallback,
				callback,
				MH_pathDropCallbackImpl,
				FD_GLFWdropfun,
				pathDropCallbackArena);
	}

	public void setScrollCallback(ScrollCallback callback) {
		scrollCallbackArena = setCallback(MH_glfwSetScrollCallback,
				window.glfw().MS_glfwSetScrollCallback,
				callback,
				MH_scrollCallbackImpl,
				FD_GLFWscrollfun,
				scrollCallbackArena);
	}

	public void setWindowPosCallback(WindowPosCallback callback) {
		windowPosCallbackArena = setCallback(MH_glfwSetWindowPosCallback,
				window.glfw().MS_glfwSetWindowPosCallback,
				callback,
				MH_windowPosCallbackImpl,
				FD_GLFWwindowposfun,
				windowPosCallbackArena);
	}

	public void setWindowSizeCallback(WindowSizeCallback callback) {
		windowSizeCallbackArena = setCallback(MH_glfwSetWindowSizeCallback,
				window.glfw().MS_glfwSetWindowSizeCallback,
				callback,
				MH_windowSizeCallbackImpl,
				FD_GLFWwindowsizefun,
				windowSizeCallbackArena);
	}

	public void setWindowCloseCallback(WindowCloseCallback callback) {
		windowCloseCallbackArena = setCallback(MH_glfwSetWindowCloseCallback,
				window.glfw().MS_glfwSetWindowCloseCallback,
				callback,
				MH_windowCloseCallbackImpl,
				FD_GLFWwindowclosefun,
				windowCloseCallbackArena);
	}

	public void setWindowRefreshCallback(WindowRefreshCallback callback) {
		windowRefreshCallbackArena = setCallback(MH_glfwSetWindowRefreshCallback,
				window.glfw().MS_glfwSetWindowRefreshCallback,
				callback,
				MH_windowRefreshCallbackImpl,
				FD_GLFWwindowrefreshfun,
				windowRefreshCallbackArena);
	}

	public void setWindowFocusCallback(WindowFocusCallback callback) {
		windowFocusCallbackArena = setCallback(MH_glfwSetWindowFocusCallback,
				window.glfw().MS_glfwSetWindowFocusCallback,
				callback,
				MH_windowFocusCallbackImpl,
				FD_GLFWwindowfocusfun,
				windowFocusCallbackArena);
	}

	public void setWindowIconifyCallback(WindowIconifyCallback callback) {
		windowIconifyCallbackArena = setCallback(MH_glfwSetWindowIconifyCallback,
				window.glfw().MS_glfwSetWindowIconifyCallback,
				callback,
				MH_windowIconifyCallbackImpl,
				FD_GLFWwindowiconifyfun,
				windowIconifyCallbackArena);
	}

	public void setWindowMaximizeCallback(WindowMaximizeCallback callback) {
		windowMaximizeCallbackArena = setCallback(MH_glfwSetWindowMaximizeCallback,
				window.glfw().MS_glfwSetWindowMaximizeCallback,
				callback,
				MH_windowMaximizeCallbackImpl,
				FD_GLFWwindowmaximizefun,
				windowMaximizeCallbackArena);
	}

	public void setFramebufferSizeCallback(FramebufferSizeCallback callback) {
		framebufferSizeCallbackArena = setCallback(MH_glfwSetFramebufferSizeCallback,
				window.glfw().MS_glfwSetFramebufferSizeCallback,
				callback,
				MH_framebufferSizeCallbackImpl,
				FD_GLFWframebuffersizefun,
				framebufferSizeCallbackArena);
	}

	public void setWindowContentScaleCallback(WindowContentScaleCallback callback) {
		windowContentScaleCallbackArena = setCallback(MH_glfwSetWindowContentScaleCallback,
				window.glfw().MS_glfwSetWindowContentScaleCallback,
				callback,
				MH_windowContentScaleCallbackImpl,
				FD_GLFWwindowcontentscalefun,
				windowContentScaleCallbackArena);
	}
}
